@@ -6,9 +6,13 @@ use std::{
 };
 
 use itertools::Itertools;
+#[cfg(feature = "serde")]
 use serde::{Deserialize, Serialize};
 
-use crate::{words::GUESSES, WordleError};
+use crate::{
+    words::GUESSES,
+    {PuzzleError, Result},
+};
 
 pub mod stupid;
 
@@ -16,7 +20,12 @@ pub mod stupid;
 ///
 /// This struct represents a possible Wordle guess, and its construction
 /// is validated to ensure that every instance is a possible word.
-#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Serialize, Deserialize, Hash)]
+#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
+#[cfg_attr(
+    feature = "serde",
+    derive(Serialize, Deserialize),
+    serde(crate = "serde_crate")
+)]
 pub struct Word {
     pub(crate) index: usize,
 }
@@ -39,11 +48,11 @@ impl Word {
     /// #
     /// # Ok::<_, wordle_rs::WordleError>(())
     /// ```
-    pub fn from_index(index: usize) -> Result<Self, WordleError> {
+    pub fn from_index(index: usize) -> Result<Self> {
         if index < GUESSES.len() {
             Ok(Word { index })
         } else {
-            Err(WordleError::InvalidIndex(index))
+            Err(PuzzleError::InvalidIndex(index).into())
         }
     }
 
@@ -65,11 +74,11 @@ impl Word {
     /// # Ok::<_, wordle_rs::WordleError>(())
     /// ```
     #[allow(clippy::should_implement_trait)]
-    pub fn from_str(word: &str) -> Result<Self, WordleError> {
+    pub fn from_str(word: &str) -> Result<Self> {
         GUESSES
             .binary_search(&word)
             .map(|index| Word { index })
-            .map_err(|_| WordleError::NotInWordlist(word.to_string()))
+            .map_err(|_| PuzzleError::NotInWordlist(word.to_string()).into())
     }
 }
 
@@ -189,11 +198,7 @@ impl Puzzle {
     /// #
     /// # Ok::<_, wordle_rs::WordleError>(())
     /// ```
-    pub fn check(
-        &mut self,
-        guess: &Word,
-        attempts: &mut Attempts,
-    ) -> Result<([Grade; 5], bool), WordleError> {
+    pub fn check(&mut self, guess: &Word, attempts: &mut Attempts) -> Result<([Grade; 5], bool)> {
         if attempts.cheat {
             self.poisoned = true;
         }
@@ -206,7 +211,7 @@ impl Puzzle {
         }
 
         if attempts.push(*guess).is_err() {
-            return Err(WordleError::OutOfGuesses);
+            return Err(PuzzleError::OutOfGuesses.into());
         }
 
         Ok(self.check_inner(guess))
@@ -252,12 +257,7 @@ impl Puzzle {
         (res, correct)
     }
 
-    fn hardmode_guard(
-        &self,
-        previous: &Word,
-        grades: &[Grade],
-        guess: &Word,
-    ) -> Result<(), WordleError> {
+    fn hardmode_guard(&self, previous: &Word, grades: &[Grade], guess: &Word) -> Result<()> {
         // We need to check that `guess` incorporates all _revealed_ guesses.
         // That means that it uses the all of the almosts and correctly uses
         // all of the corrects.
@@ -276,7 +276,7 @@ impl Puzzle {
                 Grade::Correct => {
                     // make sure prev == new since they know where this letter goes
                     if prev != new {
-                        return Err(WordleError::InvalidHardmodeGuess);
+                        return Err(PuzzleError::InvalidHardmodeGuess.into());
                     }
                 }
                 Grade::Incorrect => {}
@@ -286,7 +286,7 @@ impl Puzzle {
                     if guess.chars().filter(|&c| c == prev).count()
                         < almost_lookup[i(prev)] as usize
                     {
-                        return Err(WordleError::InvalidHardmodeGuess);
+                        return Err(PuzzleError::InvalidHardmodeGuess.into());
                     }
                 }
             }
@@ -329,6 +329,13 @@ impl AttemptsKey {
         AttemptsKey { hard, cheat: false }
     }
 
+    /// Creates a new [`AttemptsKey`] marked as a cheating key.
+    ///
+    /// Using a key created by this function to make an [`Attempts`] will mark
+    /// that instance as cheating as well, and passing that instance to
+    /// [`Puzzle::check()`] will cause the puzzle to become poisoned.
+    /// The test harness will detect this and refuse to provide performance
+    /// records after a puzzle is poisoned.
     pub fn new_cheat(hard: bool) -> AttemptsKey {
         AttemptsKey { hard, cheat: true }
     }
@@ -343,19 +350,19 @@ impl AttemptsKey {
 ///
 /// Strategies must return this, and the struct simply wraps a [`Vec`] to
 /// ensure that strategies cannot inflate their performance. While implementing
-/// [`Strategy::solve()`], you must create an instance of [`Attempts`] and
-/// pass a mutable reference to it to [`Puzzle::check()`], which updates it.
-///
-/// There are some helper methods to make stopping your algorithms possible.
+/// [`Strategy::solve()`], you must create an instance of [`Attempts`] by
+/// calling [`AttemptsKey::unlock()`] on the `key` parameter. Then,
+/// [`Puzzle::check()`] will update the attempts for you.
 ///
 /// # Examples
 ///
 /// ```rust
 /// # use std::ops::Deref;
 /// # use wordle_rs::strategy::Attempts;
-/// use wordle_rs::strategy::{Puzzle, Word};
+/// use wordle_rs::strategy::{Puzzle, Word, AttemptsKey};
 ///
-/// let mut attempts = Attempts::cheat(true);
+/// let key = AttemptsKey::new_cheat(true);
+/// let mut attempts = key.unlock();
 /// let mut puzzle = Puzzle::new(Word::from_str("limit").unwrap());
 /// let (_, _) = puzzle
 ///     .check(&Word::from_str("tithe").unwrap(), &mut attempts)
@@ -398,12 +405,12 @@ impl Attempts {
     ///
     /// This will return an error if `inner` already has six elements.
     /// Otherwise, this function will succeed.
-    pub(crate) fn push(&mut self, word: Word) -> Result<usize, Word> {
+    pub(crate) fn push(&mut self, word: Word) -> Result<usize> {
         if self.inner.len() < 6 {
             self.inner.push(word);
             Ok(self.inner.len() - 1)
         } else {
-            Err(word)
+            Err(PuzzleError::OutOfGuesses.into())
         }
     }
 
@@ -563,7 +570,95 @@ pub trait Strategy: Display + Debug + Sync {
 #[cfg(test)]
 mod test {
     use super::*;
-    use crate::WordleError;
+    use crate::{words::GUESSES, WordleError};
+
+    #[test]
+    fn word_from_index() -> Result<()> {
+        Word::from_index(GUESSES.len() - 1)?;
+        Word::from_index(5574)?;
+        Word::from_index(GUESSES.len()).unwrap_err();
+
+        Ok(())
+    }
+
+    #[test]
+    fn word_from_str() -> Result<()> {
+        assert_eq!("tithe", Word::from_str("tithe")?.deref());
+        Word::from_str("doubt")?;
+        for word in &GUESSES[4452..4460] {
+            assert_eq!(*word, Word::from_str(word)?.deref());
+        }
+        Word::from_str("aaaaa").unwrap_err();
+        Word::from_str("tithes").unwrap_err();
+
+        Ok(())
+    }
+
+    #[test]
+    fn fmt_word() {
+        assert_eq!("tithe", format!("{}", Word::from_str("tithe").unwrap()));
+        for word in &GUESSES[3141..3149] {
+            assert_eq!(*word, format!("{}", Word::from_str(word).unwrap().deref()));
+        }
+    }
+
+    #[test]
+    fn puzzle_poisoning() -> Result<()> {
+        let mut puzzle = Puzzle::new(Word::from_str("nerds")?);
+        assert!(!puzzle.poisoned);
+
+        let safe_key = AttemptsKey::new(false);
+        let mut safe_attempts = safe_key.unlock();
+
+        let (safe_grades, safe_correct) =
+            puzzle.check(&Word::from_str("doubt")?, &mut safe_attempts)?;
+        assert!(!puzzle.poisoned);
+
+        let cheat_key = AttemptsKey::new_cheat(false);
+        let mut cheat_attempts = cheat_key.unlock();
+
+        let (cheat_grades, cheat_correct) =
+            puzzle.check(&Word::from_str("doubt")?, &mut cheat_attempts)?;
+        assert!(puzzle.poisoned);
+
+        assert_eq!(safe_grades, cheat_grades);
+        assert_eq!(safe_correct, cheat_correct);
+
+        let _ = puzzle.check(&Word::from_str("gorge")?, &mut safe_attempts)?;
+        assert!(puzzle.poisoned);
+
+        Ok(())
+    }
+
+    #[test]
+    fn puzzle_out_of_guesses() -> Result<()> {
+        let mut puzzle = Puzzle::new(Word::from_str("nerds")?);
+
+        let key = AttemptsKey::new(false);
+        let mut attempts = key.unlock();
+
+        for &word in &GUESSES[100..106] {
+            let _ = puzzle.check(&Word::from_str(word)?, &mut attempts)?;
+        }
+
+        assert!(attempts.finished());
+
+        assert!(puzzle
+            .check(&Word::from_str("mount")?, &mut attempts)
+            .is_err());
+
+        assert_eq!(
+            format!(
+                "{}\n{}\n{}\n{}\n{}\n{}",
+                GUESSES[100], GUESSES[101], GUESSES[102], GUESSES[103], GUESSES[104], GUESSES[105]
+            ),
+            format!("{}", attempts)
+        );
+
+        Ok(())
+    }
+
+    // PUZZLE LOGIC TESTS
 
     fn str_to_grades(input: &str) -> [Grade; 5] {
         let mut res = [Grade::Incorrect; 5];
@@ -600,7 +695,7 @@ mod test {
 
         ($fn_name:ident[hard = $hard:expr, $answer:expr => $( [$guess:expr, $works:expr, $res:expr] );*] $other:block) => {
             #[test]
-            fn $fn_name() -> Result<(), WordleError> {
+            fn $fn_name() -> Result<()> {
                 let mut attempts = Attempts::cheat($hard);
                 let mut puzzle = Puzzle::new(Word::from_str($answer)?);
                 let mut count = 0;
